From stern@rowland.harvard.edu Tue May 31 13:47:04 2005
Date: Tue, 31 May 2005 16:33:21 -0400 (EDT)
From: Alan Stern <stern@rowland.harvard.edu>
To: Greg KH <greg@kroah.com>
Subject: USB UHCI: Detect invalid ports
Message-ID: <Pine.LNX.4.44L0.0505311628400.32520-100000@iolanthe.rowland.org>

This patch changes the way uhci-hcd detects valid ports.  The 
specification doesn't mention any way to find out how many ports a 
controller has, so the driver has to use some heuristics, reading the port 
status and control register and deciding whether the value makes sense.  
With this patch the driver will recognize a typical failure mode (all bits 
set to one) for nonexistent ports and won't assume there are always at 
least 2 ports -- such an assumption seems silly if the heuristics have 
already shown that the ports don't exist.

Signed-off-by: Alan Stern <stern@rowland.harvard.edu>
Signed-off-by: Greg Kroah-Hartman <gregkh@suse.de>

---
 drivers/usb/host/uhci-hcd.c |   14 +++++++-------
 1 files changed, 7 insertions(+), 7 deletions(-)

--- gregkh-2.6.orig/drivers/usb/host/uhci-hcd.c	2005-06-01 14:13:46.000000000 -0700
+++ gregkh-2.6/drivers/usb/host/uhci-hcd.c	2005-06-01 14:19:26.000000000 -0700
@@ -495,24 +495,24 @@
 
 	/* The UHCI spec says devices must have 2 ports, and goes on to say
 	 * they may have more but gives no way to determine how many there
-	 * are.  However, according to the UHCI spec, Bit 7 of the port
+	 * are.  However according to the UHCI spec, Bit 7 of the port
 	 * status and control register is always set to 1.  So we try to
-	 * use this to our advantage.
+	 * use this to our advantage.  Another common failure mode when
+	 * a nonexistent register is addressed is to return all ones, so
+	 * we test for that also.
 	 */
 	for (port = 0; port < (io_size - USBPORTSC1) / 2; port++) {
 		unsigned int portstatus;
 
 		portstatus = inw(uhci->io_addr + USBPORTSC1 + (port * 2));
-		if (!(portstatus & 0x0080))
+		if (!(portstatus & 0x0080) || portstatus == 0xffff)
 			break;
 	}
 	if (debug)
 		dev_info(uhci_dev(uhci), "detected %d ports\n", port);
 
-	/* Anything less than 2 or greater than 7 is weird,
-	 * so we'll ignore it.
-	 */
-	if (port < 2 || port > UHCI_RH_MAXCHILD) {
+	/* Anything greater than 7 is weird so we'll ignore it. */
+	if (port > UHCI_RH_MAXCHILD) {
 		dev_info(uhci_dev(uhci), "port count misdetected? "
 				"forcing to 2 ports\n");
 		port = 2;
